! WHIZARD 1.21 Feb 26 2002
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module particles

  use kinds, only: default !NODEP!
  use limits, only: PRT_NAME_LEN
  use lorentz
  use parameters, only: parameter_set !NODEP!

  implicit none
  private

  type, public :: particle
     private
     integer :: code
     real(kind=default) :: mass
     type(four_momentum) :: p
  end type particle

  public :: particle_new
  public :: write
  public :: operator(*)
  public :: operator(+)
  public :: set
  public :: particle_code
  public :: particle_momentum
  public :: particle_mass
  public :: particle_name
  public :: particle_tex_form
  public :: particle_width
  public :: particle_color
  public :: particle_3charge
  public :: particle_charge
  public :: particle_spin_type
  public :: particle_multiplicity
  public :: is_parton
  public :: visible
  public :: antiparticle_code

  type(particle), parameter, public :: particle_null = &
       & particle( 0, 0._default, four_momentum_null )
  integer, parameter, public :: SCALAR=1, FERMION=2, VECTOR_BOSON=3
  integer, parameter, public :: UNKNOWN=0, MASSLESS_BOSON=2
  integer, parameter, public :: &
       & D_QUARK=1, U_QUARK=2, S_QUARK=3, C_QUARK=4, B_QUARK=5, T_QUARK=6
  integer, parameter, public :: &
       & E_LEPTON=11, ELECTRON=E_LEPTON, E_NEUTRINO=12, &
       & MU_LEPTON=13, MU_NEUTRINO=14, &
       & TAU_LEPTON=15, TAU_NEUTRINO=16
  integer, parameter, public :: &
       & GLUON=21, PHOTON=22, GAMMA=PHOTON, Z_BOSON=23, W_BOSON=24, &
       & GLUON2=9
  integer, parameter, public :: &
       & HIGGS=25, &
       & LIGHT_HIGGS=HIGGS, HEAVY_HIGGS=35, AXIAL_HIGGS=36, CHARGED_HIGGS=37
  integer, parameter, public :: PROTON=2212
  integer, parameter, public :: &
       & COMPOSITE=99, UNDEFINED=0
       
  interface particle_new
     module procedure particle_new_p
     module procedure particle_new_mass
     module procedure particle_new_code
  end interface
  interface write
     module procedure particle_write_unit
  end interface
  interface operator(*)
     module procedure prod_LT_prt
     module procedure prod_prt_prt
  end interface
  interface operator(+)
     module procedure sum_prt_prt
  end interface
  interface set
     module procedure particle_set_p
     module procedure particle_set_code
  end interface
  interface particle_code
     module procedure particle_code_single, particle_code_multi
  end interface
  interface particle_momentum
     module procedure particle_momentum_multi, particle_momentum_single
  end interface
  interface particle_mass
     module procedure particle_mass_single, particle_mass_multi
  end interface
  interface particle_code
     module procedure particle_code_name
  end interface
  interface particle_mass
     module procedure particle_mass_code
  end interface
  interface particle_width
     module procedure particle_width_code
  end interface
  interface particle_color
     module procedure particle_color_code
     module procedure particle_color_get
  end interface
  interface particle_3charge
     module procedure particle_3charge_code
     module procedure particle_3charge_get
  end interface
  interface particle_charge
     module procedure particle_charge_code
     module procedure particle_charge_get
  end interface
  interface particle_spin_type
     module procedure particle_spin_type_code
  end interface
  interface particle_multiplicity
     module procedure particle_multiplicity_code
  end interface
  interface is_parton
     module procedure is_parton_multi, is_parton_single
  end interface
  interface visible
     module procedure visible_multi, visible_single
  end interface

contains

  function particle_new_p(code, mass, p) result(prt)
    integer, intent(in) :: code
    real(kind=default), optional :: mass
    type(four_momentum), intent(in) :: p
    type(particle) :: prt
    prt%code = code
    if (present(mass)) then
       prt%mass = mass
    else
       prt%mass = p**1
    end if
    prt%p    = p
  end function particle_new_p
  function particle_new_mass(code, mass) result(prt)
    integer, intent(in) :: code
    real(kind=default), intent(in) :: mass
    type(particle) :: prt
    prt%code = code
    prt%mass = mass
    prt%p    = four_momentum_at_rest(mass)
  end function particle_new_mass
  function particle_new_code(code, par) result(prt)
    integer, intent(in) :: code
    type(parameter_set), intent(in) :: par
    type(particle) :: prt
    prt = particle_new_mass( code, particle_mass(code, par) )
  end function particle_new_code

  subroutine particle_write_unit(u,prt)
    integer, intent(in) :: u
    type(particle), intent(in) :: prt
    character(len=PRT_NAME_LEN) :: prt_name
    prt_name = particle_name(prt%code)
    write(u,*) 'Particle type: ', prt_name
    write(u,*) 'M =', prt%mass
    call write(u,prt%p)
  end subroutine particle_write_unit
  function prod_LT_prt(L, prt0) result(prt)
    type(lorentz_transformation), intent(in) :: L
    type(particle), intent(in) :: prt0
    type(particle) :: prt
    prt = particle( prt0%code, prt0%mass, L * prt0%p )
  end function prod_LT_prt
  function prod_prt_prt(prt1, prt2) result(s)
    type(particle), intent(in) :: prt1, prt2
    real(kind=default) :: s
    s = prt1%p * prt2%p
  end function prod_prt_prt
    
  function sum_prt_prt(prt1, prt2) result(prt)
    type(particle), intent(in) :: prt1, prt2
    type(particle) :: prt
    prt = particle_new_p( code = COMPOSITE,  p = prt1%p + prt2%p)
  end function sum_prt_prt

  subroutine particle_set_p(prt, p)
    type(particle), intent(inout) :: prt
    type(four_momentum), intent(in) :: p
    prt%p = p
  end subroutine particle_set_p
  subroutine particle_set_code(prt, code)
    type(particle), intent(inout) :: prt
    integer, intent(in) :: code
    prt%code = code
  end subroutine particle_set_code
  function particle_code_multi (prt) result (code)
    type(particle), dimension(:), intent(in) :: prt
    integer, dimension(size(prt)) :: code
    integer :: i
    do i=1, size(prt)
       code(i) = particle_code_single (prt(i))
    end do
  end function particle_code_multi
  function particle_code_single (prt) result (code)
    type(particle), intent(in) :: prt
    integer :: code
    code = prt%code
  end function particle_code_single
  function particle_momentum_multi (prt) result (p)
    type(particle), dimension(:), intent(in) :: prt
    type(four_momentum), dimension(size(prt)) :: p
    integer :: i
    do i=1, size(prt)
       p(i) = particle_momentum_single (prt(i))
    end do
  end function particle_momentum_multi
  function particle_momentum_single (prt) result (p)
    type(particle), intent(in) :: prt
    type(four_momentum) :: p
    p = prt%p
  end function particle_momentum_single
  function particle_mass_multi (prt) result (mass)
    type(particle), dimension(:), intent(in) :: prt
    real(kind=default), dimension(size(prt)) :: mass
    integer :: i
    do i=1, size(prt)
       mass(i) = particle_mass_single (prt(i))
    end do
  end function particle_mass_multi
  function particle_mass_single (prt) result (mass)
    type(particle), intent(in) :: prt
    real(kind=default) :: mass
    mass = prt%mass
  end function particle_mass_single
  function particle_name(code) result(name)
    integer, intent(in) :: code
    character(len=PRT_NAME_LEN) :: name, n
    select case(abs(code))
    case(D_QUARK);       n = 'd'
    case(U_QUARK);       n = 'u'
    case(S_QUARK);       n = 's'
    case(C_QUARK);       n = 'c'
    case(B_QUARK);       n = 'b'
    case(T_QUARK);       n = 't'
    case(E_LEPTON);      n = 'e'
    case(E_NEUTRINO);    n = 'nu_e'
    case(MU_LEPTON);     n = 'mu'
    case(MU_NEUTRINO);   n = 'nu_mu'
    case(TAU_LEPTON);    n = 'tau'
    case(TAU_NEUTRINO);  n = 'nu_tau'
    case(GLUON,GLUON2);  n = 'g'
    case(PHOTON);        n = 'gamma'
    case(Z_BOSON);       n = 'Z'
    case(W_BOSON);       n = 'W+'
    case(HIGGS);         n = 'h'
    case(HEAVY_HIGGS);   n = 'HH'
    case(AXIAL_HIGGS);   n = 'HA'
    case(CHARGED_HIGGS); n = 'H+'
    case(-CHARGED_HIGGS);n = 'H-'
    case(PROTON);        n = 'p'
    case(COMPOSITE);     n = 'COMPOSITE'
    case(0);             n = 'UNDEFINED'
    case default;        n = "UNDEFINED"
    end select
    if (code>=0) then
       name = n
    else
       name = 'a-'//n
    end if
  end function particle_name
  function particle_tex_form(code) result(n)
    integer, intent(in) :: code
    character(len=PRT_NAME_LEN) :: n
    select case(code)
    case(D_QUARK);       n = 'd'
    case(U_QUARK);       n = 'u'
    case(S_QUARK);       n = 's'
    case(C_QUARK);       n = 'c'
    case(B_QUARK);       n = 'b'
    case(T_QUARK);       n = 't'
    case(-D_QUARK);       n = '\bar d'
    case(-U_QUARK);       n = '\bar u'
    case(-S_QUARK);       n = '\bar s'
    case(-C_QUARK);       n = '\bar c'
    case(-B_QUARK);       n = '\bar b'
    case(-T_QUARK);       n = '\bar t'
    case(E_LEPTON);      n = 'e^-'
    case(E_NEUTRINO);    n = '\nu_e'
    case(MU_LEPTON);     n = '\mu^-'
    case(MU_NEUTRINO);   n = '\nu_\mu'
    case(TAU_LEPTON);    n = '\tau^-'
    case(TAU_NEUTRINO);  n = '\nu_\tau'
    case(-E_LEPTON);      n = 'e^+'
    case(-E_NEUTRINO);    n = '\bar\nu_e'
    case(-MU_LEPTON);     n = '\mu^+'
    case(-MU_NEUTRINO);   n = '\bar\nu_\mu'
    case(-TAU_LEPTON);    n = '\tau^+'
    case(-TAU_NEUTRINO);  n = '\bar\nu_\tau'
    case(GLUON,GLUON2);  n = 'g'
    case(PHOTON);        n = '\gamma'
    case(Z_BOSON);       n = 'Z'
    case(W_BOSON);       n = 'W^+'
    case(-W_BOSON);       n = 'W^-'
    case(HIGGS);         n = 'H'
  !  case(LIGHT_HIGGS);   n = 'h'
    case(HEAVY_HIGGS);   n = 'H'
    case(AXIAL_HIGGS);   n = 'A'
    case(CHARGED_HIGGS); n = 'H^+'
    case(-CHARGED_HIGGS); n = 'H^-'
    case(PROTON);         n = 'p'
    case(-PROTON);        n = 'p~'
    case default;        n = ' '
    end select
  end function particle_tex_form
  function particle_code_name(string) result(code)
    character(len=*), intent(in) :: string
    character(len=PRT_NAME_LEN) :: name
    integer :: code
    name = string
    select case(name)
    case('d     ', 'down  ');  code = D_QUARK
    case('u     ', 'up    ');  code = U_QUARK
    case('s     ', 'strang');  code = S_QUARK
    case('c     ', 'charm ');  code = C_QUARK
    case('b     ', 'bottom');  code = B_QUARK
    case('t     ', 'top   ');  code = T_QUARK
    case('D     ', 'a-d   ', 'd~    ');  code = -D_QUARK
    case('U     ', 'a-u   ', 'u~    ');  code = -U_QUARK
    case('S     ', 'a-s   ', 's~    ');  code = -S_QUARK
    case('C     ', 'a-c   ', 'c~    ');  code = -C_QUARK
    case('B     ', 'a-b   ', 'b~    ');  code = -B_QUARK
    case('T     ', 'a-t   ', 't~    ');  code = -T_QUARK
    case('e1    ', 'e     ', 'e-    ', 'electr');  code = E_LEPTON
    case('n1    ', 'nu_e  ', 'e-neut', 've    ');  code = E_NEUTRINO
    case('e2    ', 'mu    ', 'mu-   ', 'muon  ');  code = MU_LEPTON
    case('n2    ', 'nu_mu ', 'mu-neu', 'vm    ');  code = MU_NEUTRINO
    case('e3    ', 'tau   ', 'tau-  ', 'ta-   ');  code = TAU_LEPTON
    case('n3    ', 'nu_tau', 'tau-ne', 'vt    ');  code = TAU_NEUTRINO
    case('E1    ', 'a-e   ', 'e+    ');            code = -E_LEPTON
    case('N1    ', 'a-nu_e', 've~   ');            code = -E_NEUTRINO
    case('E2    ', 'a-mu  ', 'mu+   ');            code = -MU_LEPTON
    case('N2    ', 'a-nu_m', 'vm~   ');            code = -MU_NEUTRINO
    case('E3    ', 'a-tau ', 'tau+  ', 'ta+   ');  code = -TAU_LEPTON
    case('N3    ', 'a-nu_t', 'vt~   ');            code = -TAU_NEUTRINO
    case('G     ', 'g     ', 'gluon ');            code = GLUON
    case('A     ', 'a     ', 'gamma ', 'photon');  code = PHOTON
    case('Z     ', 'z     ', 'Z-boso');            code = Z_BOSON
    case('W+    ', 'w+    ', 'W-boso');            code = W_BOSON
    case('W-    ', 'w-    ', 'a-W+  ');            code = -W_BOSON
    case('H     ', 'Higgs ', 'higgs ');  code = HIGGS
    case('h     ');  code = LIGHT_HIGGS
    case('HH    ');  code = HEAVY_HIGGS
    case('HA    ');  code = AXIAL_HIGGS
    case('H+    ');  code = CHARGED_HIGGS
    case('H-    ');  code = -CHARGED_HIGGS
    case('p     ');            code = PROTON
    case('a-p   ', 'p~    ');  code = -PROTON
    case default;    code = UNDEFINED
    end select
  end function particle_code_name

  function particle_mass_code(code, par) result(m)
    integer, intent(in) :: code
    type(parameter_set), intent(in) :: par
    real(kind=default) :: m
    select case(abs(code))
    case(S_QUARK);       m = par%ms
    case(C_QUARK);       m = par%mc
    case(B_QUARK);       m = par%mb
    case(T_QUARK);       m = par%mtop 
    case(E_LEPTON);      m = par%me
    case(MU_LEPTON);     m = par%mm
    case(TAU_LEPTON);    m = par%mt
    case(Z_BOSON);       m = par%mZ
    case(W_BOSON);       m = par%mW
    case(HIGGS);         m = par%mh
  !   case(HEAVY_HIGGS);   m = par%mHH
  !   case(AXIAL_HIGGS);   m = par%mHA
  !   case(CHARGED_HIGGS); m = par%mHpm
    case default;        m = 0.
    end select
  end function particle_mass_code
  function particle_width_code(code, par) result(w)
    integer, intent(in) :: code
    type(parameter_set), intent(in) :: par
    real(kind=default) :: w
    select case(abs(code))
    case(T_QUARK);       w = par%wtop
    case(Z_BOSON);       w = par%wZ
    case(W_BOSON);       w = par%wW
    case(HIGGS);         w = par%wh
  !   case(HEAVY_HIGGS);   w = par%wHH
  !   case(AXIAL_HIGGS);   w = par%wHA
  !   case(CHARGED_HIGGS); w = par%wHpm
    case default;        w = 0.
    end select
  end function particle_width_code
  function particle_color_code(code) result (Q)
    integer, intent(in) :: code
    integer :: Q
    select case(abs(code))
    case( D_QUARK, U_QUARK, S_QUARK, C_QUARK, B_QUARK, T_QUARK);  Q = 3
    case(-D_QUARK,-U_QUARK,-S_QUARK,-C_QUARK,-B_QUARK,-T_QUARK);  Q = -3
    case(GLUON,GLUON2);      Q = 8
    case default;        Q = 0
    end select
    if (code<0) Q = -Q
  end function particle_color_code
  function particle_color_get(prt) result (Q)
    type(particle), intent(in) :: prt
    integer :: Q
    Q = particle_color_code(particle_code(prt))
  end function particle_color_get

  function particle_3charge_code(code) result (Q)
    integer, intent(in) :: code
    integer :: Q
    select case(abs(code))
    case(D_QUARK);       Q = -1
    case(U_QUARK);       Q = 2
    case(S_QUARK);       Q = -1
    case(C_QUARK);       Q = 2
    case(B_QUARK);       Q = -1
    case(T_QUARK);       Q = 2
    case(E_LEPTON);      Q = -3
    case(MU_LEPTON);     Q = -3
    case(TAU_LEPTON);    Q = -3
    case(W_BOSON);       Q = 3
    case(CHARGED_HIGGS); Q = 3
    case default;        Q = 0
    end select
    if (code<0) Q = -Q
  end function particle_3charge_code
  function particle_3charge_get(prt) result (Q)
    type(particle), intent(in) :: prt
    integer :: Q
    Q = particle_3charge_code(particle_code(prt))
  end function particle_3charge_get

  function particle_charge_code(code) result (Q)
    integer, intent(in) :: code
    real(kind=default) :: Q
    Q = particle_3charge_code(code) / 3._default
  end function particle_charge_code
  function particle_charge_get(prt) result (Q)
    type(particle), intent(in) :: prt
    real(kind=default) :: Q
    Q = particle_charge_code(particle_code(prt))
  end function particle_charge_get

  function particle_spin_type_code(code) result (spin)
    integer, intent(in) :: code
    integer :: spin
    select case(abs(code))
    case(D_QUARK, U_QUARK, S_QUARK, C_QUARK, B_QUARK, T_QUARK)
       spin = FERMION
    case(E_LEPTON, E_NEUTRINO, MU_LEPTON, MU_NEUTRINO, &
         & TAU_LEPTON, TAU_NEUTRINO)
       spin = FERMION
    case(GLUON, GLUON2, PHOTON, Z_BOSON, W_BOSON)
       spin = VECTOR_BOSON
    case(HIGGS, HEAVY_HIGGS, AXIAL_HIGGS, CHARGED_HIGGS)
       spin = SCALAR
    case(PROTON)
       spin = FERMION
    case default
       spin = UNKNOWN
    end select
  end function particle_spin_type_code

  function particle_multiplicity_code (code) result (multiplicity)
    integer, intent(in) :: code
    integer :: multiplicity
    select case (abs(code))
    case(GLUON, GLUON2, PHOTON)
       multiplicity = MASSLESS_BOSON
    case default
       multiplicity = particle_spin_type (code)
    end select
  end function particle_multiplicity_code

  function is_parton_multi (code) result (ok)
    integer, dimension(:), intent(in) :: code
    logical, dimension(size(code)) :: ok
    integer :: i
    do i=1, size(code)
       ok(i) = is_parton_single (code(i))
    end do
  end function is_parton_multi

  function is_parton_single (code) result (ok)
    integer, intent(in) :: code
    logical :: ok
    select case (abs(code))
    case(D_QUARK, U_QUARK, S_QUARK, C_QUARK, B_QUARK, T_QUARK)
       ok = .true.
    case(GLUON, GLUON2)
       ok = .true.
    case default
       ok = .false.
    end select
  end function is_parton_single

  function visible_multi (code) result (ok)
    integer, dimension(:), intent(in) :: code
    logical, dimension(size(code)) :: ok
    integer :: i
    do i=1, size(code)
       ok(i) = visible_single (code(i))
    end do
  end function visible_multi

  function visible_single (code) result (ok)
    integer, intent(in) :: code
    logical :: ok
    ok = (particle_3charge(code) /= 0) .or. (particle_color(code) /= 0)
  end function visible_single

  function antiparticle_code (code) result (acode)
    integer, intent(in) :: code
    integer :: acode
    logical :: antiparticle
    select case(abs(code))
    case(D_QUARK, U_QUARK, S_QUARK, C_QUARK, B_QUARK, T_QUARK)
       antiparticle = .true.
    case(E_LEPTON, E_NEUTRINO, MU_LEPTON, MU_NEUTRINO, TAU_LEPTON, TAU_NEUTRINO)
       antiparticle = .true.
    case(GLUON, GLUON2, PHOTON, Z_BOSON)
       antiparticle = .false.
    case(W_BOSON)
       antiparticle = .true.
    case(HIGGS, HEAVY_HIGGS, AXIAL_HIGGS)
       antiparticle = .false.
    case(CHARGED_HIGGS)
       antiparticle = .true.
    case(PROTON)
       antiparticle = .true.
    case default
       antiparticle = .false.
    end select
    if (antiparticle) then
       acode = -code
    else
       acode = code
    end if
  end function antiparticle_code


end module particles
