! WHIZARD 1.21 Feb 26 2002
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module phase_space

  use kinds, only: default !NODEP!
  use lorentz
  use constants
  use particles
  use decay_trees, only: decay_tree
  use mappings
  use decay_forests
  use events
!  use vamp_rest, only: division, vamp_grid, vamp_grids !NODEP!

  implicit none
  private

  public :: event_fill
  public :: map_decay_tree_using_grid

contains

  subroutine event_fill(evt, r, f, factor, volume, chan, active, ok)
    type(event), intent(inout) :: evt
    real(kind=default), dimension(:,:), intent(inout) :: r
    type(decay_forest), intent(in) :: f
    real(kind=default), dimension(:), intent(out) :: factor
    real(kind=default), intent(out) :: volume
    integer, intent(in) :: chan
    logical, dimension(:), intent(in) :: active
    logical, intent(out) :: ok
    type(particle), dimension(:), allocatable :: prt
    integer :: i,n1,n2
    n1 = f%n_dim1
    n2 = f%n_dim2
    allocate(prt(f%n_branches_tot))
    call setup_externals(prt, evt%prt(1:))
    if (evt%scattering) then
       do i=1,2
          prt(2**(f%n_externals-i)) = evt%prt(-i)
       end do
       prt(2**(f%n_externals-2)-1)  = evt%prt(0)
    else
       prt(2**(f%n_externals-1)-1)  = evt%prt(0)
    end if
    call fill_decay_tree(prt, &
         &               factor(chan), volume, &
         &               f%tree(chan), f%map(:,chan), f%mass_sum, &
         &               evt%sqrts_hat, evt%L0, &
         &               r(1:n1,chan), r(n1+1:n1+n2,chan), ok)
    if (.not.ok) return
    do i=1, f%n_channels
       if (i/=chan .and. active(i)) then
          call combine_particles(prt, f%tree(i))
          call extract_decay_tree(prt, &
               &              factor(i), &
               &              f%tree(i), f%map(:,i), f%mass_sum, &
               &              evt%sqrts_hat, inverse(evt%L0), &
               &              r(1:n1,i), r(n1+1:n1+n2,i))
       end if
    end do
    call extract_externals(prt, evt%prt(1:))
    deallocate(prt)
  end subroutine event_fill

  subroutine setup_externals(prt, external)
    type(particle), dimension(:), intent(out) :: prt
    type(particle), dimension(:), intent(in) :: external
    integer :: i
    prt = particle_null
    do i=1, ubound(external,1)
       prt(ibset(0,i-1)) = external(i)
    end do
  end subroutine setup_externals

  subroutine extract_externals(prt, external)
    type(particle), dimension(:), intent(in) :: prt
    type(particle), dimension(1:), intent(out) :: external
    integer :: i
    do i=1, ubound(external,1)
       external(i) = prt(ibset(0,i-1))
    end do
  end subroutine extract_externals

  subroutine fill_decay_tree(prt, factor, volume, &
       &                      t, map, mass_sum, sqrts, L0, r1, r2, ok)
    type(particle), dimension(:), intent(inout) :: prt
    real(kind=default), intent(out) :: factor, volume
    type(decay_tree), intent(in) :: t
    type(mapping), dimension(:), intent(in) :: map
    real(kind=default), dimension(:), intent(in) :: mass_sum
    real(kind=default), intent(in) :: sqrts
    type(lorentz_transformation), intent(in) :: L0
    real(kind=default), dimension(:), intent(in) :: r1, r2
    logical, intent(out) :: ok
    real(kind=default), dimension(t%mask_out) :: decay_p
    call set_msq(prt, factor, volume, decay_p, t, map, mass_sum, sqrts, r1, ok)
    if (ok) call set_angles(prt, factor, decay_p, sqrts, L0, t, map, r2)
  end subroutine fill_decay_tree

  subroutine set_msq(prt, factor, volume, decay_p, t, map, &
       &             mass_sum, sqrts, r, ok)
    type(particle), dimension(:), intent(inout) :: prt
    real(kind=default), intent(out) :: factor, volume
    real(kind=default), dimension(:), intent(out) :: decay_p
    type(decay_tree), intent(in) :: t
    type(mapping), dimension(:), intent(in) :: map
    real(kind=default), dimension(:), intent(in) :: mass_sum
    real(kind=default), intent(in) :: sqrts
    real(kind=default), dimension(:), intent(in) :: r
    logical, intent(out) :: ok
    integer :: ix
    integer(kind=TC) :: k
    real(kind=default) :: m_tot, m1, m2
    ok =.true.
    ix = 1
    k  = t%mask_out
    m_tot = mass_sum(k)
    decay_p(k) = 0.
    if (m_tot < sqrts) then
       call set_msq_x(k, factor, volume, .true.)
    else
       ok = .false.
!        write(*,*) 'sqrt(s)  =', sqrts
!        write(*,*) 'mass sum =', m_tot
!        stop 'Total energy too low'
    end if
  contains
    recursive subroutine set_msq_x(k, factor, volume, initial)
      integer(kind=TC), intent(in) :: k
      real(kind=default), intent(out) :: factor, volume
      logical, intent(in) :: initial
      real(kind=default) :: msq, m, m_min, m_max, lda, rlda
      integer(kind=TC) :: k1,k2
      real(kind=default) :: f1,f2,v1,v2
      k1 = t%branch(k)%daughter(1);  k2 = t%branch(k)%daughter(2)
      if (k1==0) then
         factor = 1;  volume = 1
      else
         call set_msq_x(k1, f1, v1, .false.);  if (.not.ok) return
         call set_msq_x(k2, f2, v2, .false.);  if (.not.ok) return
         m_min = mass_sum(k)
         m_max = sqrts - m_tot + m_min
         if (initial) then
            msq = sqrts**2
            m = sqrt(msq)
            factor = f1 * f2
            volume = v1 * v2 / (4*twopi5)
         else
            call apply(map(k), r(ix), sqrts**2, &
                 &     msq, m_min**2, m_max**2, factor); ix=ix+1
            if (msq>=0) then
               m = sqrt(msq)
               factor = f1 * f2 * factor
               volume = v1 * v2 * sqrts**2 /(4*twopi2)
               prt(k) = particle_new( COMPOSITE, m )
            else
               ok = .false.
            end if
         end if
         if (ok) then
            m1 = particle_mass(prt(k1));  m2 = particle_mass(prt(k2))
            lda = lambda(msq, m1**2, m2**2)
            if (lda>0 .and. m>m1+m2 .and. m<=m_max) then
               rlda = sqrt(lda)
               decay_p(k1) = rlda/(2*m)
               decay_p(k2) = -decay_p(k1)
               factor = rlda/msq * factor
            else
               ok = .false.
            end if
         end if
      end if
    end subroutine set_msq_x
  end subroutine set_msq

  subroutine set_angles(prt, factor, decay_p, sqrts, L0, t, map, r)
    type(particle), dimension(:), intent(inout) :: prt
    real(kind=default), intent(inout) :: factor
    real(kind=default), dimension(:), intent(in) :: decay_p
    real(kind=default), intent(in) :: sqrts
    type(lorentz_transformation), intent(in) :: L0
    type(decay_tree), intent(in) :: t
    type(mapping), dimension(:), intent(in) :: map
    real(kind=default), dimension(:), intent(in) :: r
    integer :: ix
    integer(kind=TC) :: k
    ix = 1
    k  = t%mask_out
    call set_angles_x(k, L0)
  contains
    recursive subroutine set_angles_x(k, L0)
      integer(kind=TC), intent(in) :: k
      type(lorentz_transformation), intent(in) :: L0
      real(kind=default) :: m, msq, ct,st,phi,f, E,p
      type(lorentz_transformation) :: L
      integer(kind=TC) :: k1,k2
      type(three_momentum) :: axis
      p = decay_p(k)
      m = particle_mass(prt(k));  msq = m**2
      E = sqrt(msq + p**2)
      call set(prt(k), L0 * four_momentum_moving(E,p,3) )
      k1 = t%branch(k)%daughter(1);  k2 = t%branch(k)%daughter(2)
      if (k1/=0) then
         if (m>0) then
            L = L0 * boost( p/m, 3)
         else
            L = L0
         end if
         phi = r(ix)*twopi;  ix=ix+1
         call apply(map(k), prt, decay_p, k1, k2, sqrts, r(ix), ct, st, f)
         ix=ix+1
         factor = factor * f
         if (t%branch(k)%friend==0) then
            L = L * rotation(phi,3) * rotation(ct,st,2)
         else
            axis = space_part( &
                 & inverse(L) * particle_momentum(prt(t%branch(k)%friend)) )
            L = L * rotation(three_momentum_canonical(3), axis) &
                 & * rotation(phi,3) * rotation(ct,st,2)
         end if
         call set_angles_x(k1, L)
         call set_angles_x(k2, L)
      end if
    end subroutine set_angles_x
  end subroutine set_angles

  subroutine combine_particles(prt, t)
    type(particle), dimension(:), intent(inout) :: prt
    type(decay_tree), intent(in) :: t
    call combine_particles_x(t%mask_out)
  contains
    recursive subroutine combine_particles_x(k)
      integer(kind=TC), intent(in) :: k
      integer(kind=TC) :: k1,k2
      k1 = t%branch(k)%daughter(1);  k2 = t%branch(k)%daughter(2)
      if (k1/=0) then
         call combine_particles_x(k1)
         call combine_particles_x(k2)
      end if
      if (particle_code(prt(k))==0) then
         prt(k) = prt(k1) + prt(k2)
      end if
    end subroutine combine_particles_x
  end subroutine combine_particles

  subroutine extract_decay_tree(prt, factor, &
       &                        t, map, mass_sum, sqrts, L0, r1, r2)
    type(particle), dimension(:), intent(in) :: prt
    real(kind=default), intent(out) :: factor
    type(decay_tree), intent(in) :: t
    type(mapping), dimension(:), intent(in) :: map
    real(kind=default), dimension(:), intent(in) :: mass_sum
    real(kind=default), intent(in) :: sqrts
    type(lorentz_transformation), intent(in) :: L0
    real(kind=default), dimension(:), intent(out) :: r1, r2
    real(kind=default), dimension(t%mask_out) :: decay_p
    call get_msq (prt, factor, decay_p, t, map, mass_sum, sqrts, r1)
    call get_angles (prt, factor, decay_p, sqrts, L0, t, map, r2)
  end subroutine extract_decay_tree

  subroutine get_msq (prt, factor, decay_p, t, map, mass_sum, sqrts, r)
    type(particle), dimension(:), intent(in) :: prt
    real(kind=default), intent(out) :: factor
    real(kind=default), dimension(:), intent(out) :: decay_p
    type(decay_tree), intent(in) :: t
    type(mapping), dimension(:), intent(in) :: map
    real(kind=default), dimension(:), intent(in) :: mass_sum
    real(kind=default), intent(in) :: sqrts
    real(kind=default), dimension(:), intent(inout) :: r
    integer :: ix
    integer(kind=TC) :: k
    real(kind=default) :: m_tot
    ix = 1
    k  = t%mask_out
    m_tot = mass_sum(k)
    decay_p(k) = 0.
    call get_msq_x(k, factor, .true.)
  contains
    recursive subroutine get_msq_x (k, factor, initial)
      integer(kind=TC), intent(in) :: k
      real(kind=default), intent(out) :: factor
      logical, intent(in) :: initial
      real(kind=default) :: msq, m, m_min, m_max, lda, rlda
      integer(kind=TC) :: k1,k2
      real(kind=default) :: f1,f2, m1,m2
      k1 = t%branch(k)%daughter(1);  k2 = t%branch(k)%daughter(2)
      if (k1==0) then
         factor = 1
      else
         call get_msq_x (k1, f1, .false.);
         call get_msq_x (k2, f2, .false.);
         m_min = mass_sum(k)
         m_max = sqrts - m_tot + m_min
         m = particle_mass(prt(k));  msq = m**2
         if (initial) then
            factor = f1 * f2
         else
            call get(map(k), r(ix), sqrts**2, &
                 &   msq, m_min**2, m_max**2, factor);  ix=ix+1
            factor = f1 * f2 * factor
         end if
         m1 = particle_mass(prt(k1));  m2 = particle_mass(prt(k2))  
         lda = lambda(msq, m1**2, m2**2)
         if (lda>0) then
            rlda = sqrt(lda)
            decay_p(k1) = rlda/(2*m)
            decay_p(k2) = -decay_p(k1)
            factor = rlda/msq * factor
         else
            decay_p(k1) = 0
            decay_p(k2) = 0
            factor = 0
         end if
      end if
    end subroutine get_msq_x
  end subroutine get_msq

  subroutine get_angles (prt, factor, decay_p, sqrts, L0, t, map, r)
    type(particle), dimension(:), intent(in) :: prt
    real(kind=default), intent(inout) :: factor
    real(kind=default), dimension(:), intent(in) :: decay_p
    real(kind=default), intent(in) :: sqrts
    type(lorentz_transformation), intent(in) :: L0
    type(decay_tree), intent(in) :: t
    type(mapping), dimension(:), intent(in) :: map
    real(kind=default), dimension(:), intent(out) :: r
    integer :: ix
    integer(kind=TC) :: k
    ix = 1
    k  = t%mask_out
    call get_angles_x (k, L0)
  contains
    recursive subroutine get_angles_x (k, L0)
      integer(kind=TC), intent(in) :: k
      type(lorentz_transformation), intent(in) :: L0
      real(kind=default) :: m,ct,st,phi,p,f
      type(lorentz_transformation) :: L
      type(three_momentum) :: n, axis
      integer(kind=TC) :: k1,k2
      p = decay_p(k)
      k1 = t%branch(k)%daughter(1);  k2 = t%branch(k)%daughter(2)
      m = particle_mass(prt(k))
      if (k1/=0) then
         if (m>0) then
            L = boost(-p/m, 3) * L0
         else
            L = L0
         end if
         if (t%branch(k)%friend/=0) then
            axis = space_part(L * particle_momentum(prt(t%branch(k)%friend)))
            L = rotation(axis, three_momentum_canonical(3)) * L
         end if
         if (decay_p(k1) > 0) then
            n = space_part(L * particle_momentum(prt(k1))) / decay_p(k1)
         else
            n = three_momentum_null
         end if
         phi = azimuthal_angle(n)
         r(ix) = phi/twopi;  ix=ix+1
         ct = polar_angle_ct(n);  st = sqrt(1-ct**2)
         call get(map(k), prt, decay_p, k1, k2, sqrts, r(ix), ct, f);  ix=ix+1
         factor = factor * f
         L = rotation(ct,-st,2) * rotation(-phi,3) * L
         call get_angles_x(k1, L)
         call get_angles_x(k2, L)
      end if
    end subroutine get_angles_x
  end subroutine get_angles

  subroutine map_decay_tree_using_grid (map, tree, n1, n2, x, nbins)
    type(mapping), dimension(:), intent(inout) :: map
    type(decay_tree), intent(in) :: tree
    integer, intent(in) :: n1, n2
!    type(vamp_grid), intent(inout) :: g
    real(kind=default), dimension(0:,:), intent(inout) :: x
    integer, dimension(:), intent(in) :: nbins
!    if (n1>0) call map_msq_using_grid (map, tree, g%div(1:n1))
!    if (n2>0) call map_angles_using_grid (map, tree, g%div(n1+1:n1+n2))
    if (n1>0) then
       call map_msq_using_grid (map, tree, x(0:, 1:n1), nbins(1:n1))
    end if
    if (n2>0) then
       call map_angles_using_grid &
            & (map, tree, x(0:, n1+1:n1+n2), nbins(n1+1:n1+n2))
    end if
  end subroutine map_decay_tree_using_grid

  subroutine map_msq_using_grid (map, tree, x, nbins)
    type(mapping), dimension(:), intent(inout) :: map
    type(decay_tree), intent(in) :: tree
!    type(division), dimension(:), intent(inout) :: div
    real(kind=default), dimension(0:,:), intent(inout) :: x
    integer, dimension(:), intent(in) :: nbins
    integer :: ix
    ix = 1
!    print *, 'msq divisions'
    call map_msq_x (tree%mask_out, .true.)
  contains
    recursive subroutine map_msq_x (k, initial)
      integer(kind=TC), intent(in) :: k
      logical, intent(in) :: initial
      integer(kind=TC) :: k1, k2
      k1 = tree%branch(k)%daughter(1);  k2 = tree%branch(k)%daughter(2)
      if (k1/=0) then
         call map_msq_x (k1, .false.)
         call map_msq_x (k2, .false.)
         if (.not.initial) then
!            print *, 'division #', ix
!            call map_msq_using_division (map(k), div(ix));  ix = ix+1
            call map_msq_using_division (map(k), x(0:nbins(ix),ix));  ix = ix+1
         end if
      end if
    end subroutine map_msq_x
  end subroutine map_msq_using_grid

  subroutine map_angles_using_grid (map, tree, x, nbins)
    type(mapping), dimension(:), intent(inout) :: map
    type(decay_tree), intent(in) :: tree
!    type(division), dimension(:), intent(inout) :: div
    real(kind=default), dimension(0:,:), intent(inout) :: x
    integer, dimension(:), intent(in) :: nbins
    integer :: ix
    ix = 1
!    print *, 'angle divisions'
    call map_angles_x (tree%mask_out)
  contains
    recursive subroutine map_angles_x (k)
      integer(kind=TC), intent(in) :: k
      integer(kind=TC) :: k1,k2 
      k1 = tree%branch(k)%daughter(1);  k2 = tree%branch(k)%daughter(2)
      ix = ix+1  ! Skipping phi
      if (k1/=0) then
!         print *, 'division #', ix
!         call map_ct_using_division (map(k), div(ix));  ix = ix+1
         call map_ct_using_division (map(k), x(0:nbins(ix),ix));  ix = ix+1
         call map_angles_x (k1)
         call map_angles_x (k2)
      end if
    end subroutine map_angles_x
  end subroutine map_angles_using_grid


end module phase_space
